#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <math.h>
#include "trackball.h"

struct POINT {
    GLfloat x;
    GLfloat y;
    GLfloat z;
};

struct POINT CtrlPoints[4][4];
struct POINT Grid[17][17];	/* Create a 16x16 grid */

GLfloat B[4][4] = {
        { -1.0,  3.0, -3.0, 1.0},
        {  3.0, -6.0,  3.0, 0.0},       
        { -3.0,  3.0,  0.0, 0.0},
        {  1.0,  0.0,  0.0, 0.0}};

GLfloat BT[4][4] = {
        { -1.0,  3.0, -3.0, 1.0},
        {  3.0, -6.0,  3.0, 0.0},       
        { -3.0,  3.0,  0.0, 0.0},
        {  1.0,  0.0,  0.0, 0.0}};

GLfloat BP_X[4][4];
GLfloat BP_Y[4][4];
GLfloat BP_Z[4][4];

GLfloat BPB_X[4][4];
GLfloat BPB_Y[4][4];
GLfloat BPB_Z[4][4];

int in_box = -1;
int off_x, off_y;

extern GLfloat zoom;
extern GLfloat tbx;
extern GLfloat tby;


GLint mouse_state = -1;
GLint mouse_button = -1;
void Find_Nearest(int , int );
int ni, nj;
int check=0;

int target_point=0;

float max_x, max_y, max_z, min_x, min_y, min_z;
int width;
int SELECTED = GL_FALSE;
GLdouble wx[4][4], wy[4][4], wz[4][4];

static GLboolean invert_matrix( const GLfloat *m, GLfloat *out );
void find_matrix();
void move_ctrlpoint(int x, int y);

void init()
{
    float tmp;
    int i, j, k;
    FILE *inf;

    tbInit(GLUT_MIDDLE_BUTTON);

    inf = fopen("patch.dat", "r");
    if(inf==NULL) {
	printf("Error in open file!\n");
	exit(0);
    }

    for(k=0; k<16; k++) {
        i=k/4;
        j=k%4;
        fscanf(inf, "%f %f %f %f", &tmp,
        &(CtrlPoints[i][j].x), &(CtrlPoints[i][j].y), &(CtrlPoints[i][j].z));

    }

    for(k=0; k<16; k++) {
        i=k/4;
        j=k%4;
        printf("%7.4f %7.4f %7.4f\n",
                CtrlPoints[i][j].x,  CtrlPoints[i][j].y, CtrlPoints[i][j].z);
    }

    fclose(inf);

    find_matrix();
}
void find_matrix()
{
    int i, j, k;
    for(i=0; i<4; i++)
        for(j=0; j<4; j++) {
            BP_X[i][j] = 0.0;
            BP_Y[i][j] = 0.0;
            BP_Z[i][j] = 0.0;
            for(k=0; k < 4; k++) {
                BP_X[i][j] += B[i][k] * CtrlPoints[k][j].x;
                BP_Y[i][j] += B[i][k] * CtrlPoints[k][j].y;
                BP_Z[i][j] += B[i][k] * CtrlPoints[k][j].z;
            }
        }

    for(i=0; i<4; i++)
        for(j=0; j<4; j++) {
            BPB_X[i][j] = 0.0;
            BPB_Y[i][j] = 0.0;
            BPB_Z[i][j] = 0.0;
            for(k=0; k < 4; k++) {
                BPB_X[i][j] += BP_X[i][k] * BT[k][j];
                BPB_Y[i][j] += BP_Y[i][k] * BT[k][j];
                BPB_Z[i][j] += BP_Z[i][k] * BT[k][j];
            }
        }
}

void draw_Ctrl_Points()
{
    int i,j;
    glColor3f(1.0, 1.0, 1.0);

    glPointSize(3.0);

    glBegin(GL_POINTS);
    for(i=0; i<4; i++) {
	for(j=0; j<4; j++)
            glVertex3f(CtrlPoints[i][j].x, CtrlPoints[i][j].y,
			CtrlPoints[i][j].z);
    }
    glEnd();


    glPointSize(2.0);
    for(i=0; i<4; i++) {
	glBegin(GL_LINE_STRIP);
	for(j=0; j<4; j++)
            glVertex3f(CtrlPoints[i][j].x, CtrlPoints[i][j].y,
			CtrlPoints[i][j].z);
        glEnd();
    }

    for(j=0; j<4; j++) {
	glBegin(GL_LINE_STRIP);
        for(i=0; i<4; i++) 
            glVertex3f(CtrlPoints[i][j].x, CtrlPoints[i][j].y,
			CtrlPoints[i][j].z);
        glEnd();
    }

    glEnd();

    if(SELECTED) {
/*
    i=target_point/4;
    j=target_point%4;
*/
    glPointSize(5.0);
    glColor3f(1.0, 0.0, 0.0);
    glBegin(GL_POINTS);
    glVertex3f(CtrlPoints[ni][nj].x, CtrlPoints[ni][nj].y,
	CtrlPoints[ni][nj].z);
    glEnd();
    }
}

void print_grid()
{
    int i, j;
    for(i=0; i<16; i++) {
	for(j=0; j<16; j++) 
	    printf("(%4.2f, %4.2f, %4.2f) ", Grid[i][j].x, Grid[i][j].y,
		Grid[i][j].z );
	printf("\n");
    }
}

void draw_curve()
{
    int i, j, m, n;
    float u, v;
    float xx, yy, zz;
    float tx[4], ty[4], tz[4];
    GLfloat uu[4], vv[4];

    u = 0.0;
    v = 0.0;

    for(i=0; i<=16; i++) {
	u = (float)i*.0625;
        for(j=0; j<=16; j++) {
	    v = (float)j*.0625;

            uu[3] = 1.;
	    uu[2] = u;
            uu[1] = uu[2]*u;
            uu[0] = uu[1]*u;

            vv[3] = 1.;
	    vv[2] = v;
            vv[1] = vv[2]*v;
            vv[0] = vv[1]*v;

	 
	    for(n=0; n<4; n++) {
		tx[n]=0.0;
		ty[n]=0.0;
		tz[n]=0.0;
                for(m=0; m < 4; m++)  {
		    tx[n]+=uu[m]*BPB_X[m][n];
		    ty[n]+=uu[m]*BPB_Y[m][n];
		    tz[n]+=uu[m]*BPB_Z[m][n];
		}
	    }

	    xx = yy = zz = 0.0;
	
	    for(m=0; m<4; m++) {
		xx+=tx[m]*vv[m];
		yy+=ty[m]*vv[m];
		zz+=tz[m]*vv[m];
	    }
	    Grid[i][j].x = xx;
	    Grid[i][j].y = yy;
	    Grid[i][j].z = zz;
  	}
    }

/*
print_grid();
*/
    glColor3f(1.0, 1.0, 0.0);
    glPointSize(2.0);
    for(i=0; i<=16; i++) {
        glBegin(GL_LINE_STRIP);
        for(j=0; j<=16; j++)
            glVertex3f(Grid[i][j].x, Grid[i][j].y, Grid[i][j].z);
        glEnd();
    }

    for(j=0; j<=16; j++) {
        glBegin(GL_LINE_STRIP);
        for(i=0; i<=16; i++)
            glVertex3f(Grid[i][j].x, Grid[i][j].y, Grid[i][j].z);
        glEnd();
    }

}
void display(void)
{
    glClear (GL_COLOR_BUFFER_BIT);
    glPushMatrix();
    tbMatrix();

    draw_Ctrl_Points();
    draw_curve();
    glPopMatrix();
    glFlush();
}

void reshape(int w, int h)
{

    width = w;
    tbReshape(w, h);

    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 1., 30.0);
/*
    glFrustum(-4., 4., -4., 4., 10, 30.0);
*/
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0.0, 0.0, -10.0+zoom);

}

void move_ctrlpoint(int x, int y)
{
    GLdouble xx, yy, zz;
    GLdouble wld_x, wld_y, wld_z;	/* world coordinate */
    GLdouble tx, ty, tz;	
    GLint viewport[4];
    GLdouble mvmatrix[16], projmatrix[16];
    GLfloat invert_tb[16], tm[16];
    int i, j, k;

    glGetIntegerv (GL_VIEWPORT, viewport);
    glGetDoublev (GL_MODELVIEW_MATRIX, mvmatrix);
    glGetDoublev (GL_PROJECTION_MATRIX, projmatrix);


    xx = (GLdouble)x;
    yy = (GLdouble)width-(GLdouble)y;
    zz = wz[ni][nj];

    gluUnProject(xx, yy, zz,
		mvmatrix, projmatrix, viewport,
		&tx, &ty, &tz);

    k=0; 
    for(i=0; i<4; i++)
	for(j=0; j<4; j++)
	    tm[k++]=tb_transform[i][j];

    if(!invert_matrix(tm, invert_tb)) {
	printf("Erron in inverting transform matrix\n");
	exit(0);
    }

    wld_x = tx * invert_tb[0] +
	    ty * invert_tb[4] +
	    tz * invert_tb[8] +
	    invert_tb[12];

    wld_y = tx * invert_tb[1] +
	    ty * invert_tb[5] +
	    tz * invert_tb[9] +
	    invert_tb[13];

    wld_z = tx * invert_tb[2] +
	    ty * invert_tb[6] +
	    tz * invert_tb[10] +
	    invert_tb[14];

    CtrlPoints[ni][nj].x = wld_x;
    CtrlPoints[ni][nj].y = wld_y;
    CtrlPoints[ni][nj].z = wld_z;

    find_matrix();
    
}

void mouse(int button, int state, int x, int y)
{

    mouse_state = state;
    mouse_button = button;

  
    if(button == GLUT_LEFT_BUTTON && state == GLUT_UP) { 
/*
        Find_Nearest(x, y);
*/
	SELECTED = GL_FALSE;
    }
    else 
	tbMouse(button, state, x, y);
    glutPostRedisplay();


}

void keyboard(unsigned char key, int x, int y)
{
    int i, j;
    switch (key) {
      	case 27:
            exit(0);
            break;
	case 'n':
	    target_point = (target_point + 1)%16;
	    break;
	case 'x':
	    if(SELECTED)
	        CtrlPoints[ni][nj].x+=.1;
	    break;
	case 'X':
	    if(SELECTED)
	        CtrlPoints[ni][nj].x-=.1;
	    break;
	case 'y':
	    if(SELECTED)
	        CtrlPoints[ni][nj].y+=.1;
	    break;
	case 'Y':
	    if(SELECTED)
	        CtrlPoints[ni][nj].y-=.1;
	    break;
	case 'z':
	    if(SELECTED)
	        CtrlPoints[ni][nj].z+=.1;
	    break;
	case 'Z':
	    if(SELECTED)
	        CtrlPoints[ni][nj].z-=.1;
	    break;
    }
    find_matrix();
    glutPostRedisplay();
}
void
special(int key, int x, int y)
{
    switch (key) {
	case GLUT_KEY_RIGHT:
	    translate(.05, 0);
	    break;	
	case GLUT_KEY_LEFT:
	    translate(-.05, 0);
	    break;	
	case GLUT_KEY_UP:
	    translate(.0, 0.05);
	    break;	
	case GLUT_KEY_DOWN:
	    translate(.0, -0.05);
	    break;	
   }
   glutPostRedisplay();
}

void
motion(int x, int y)
{
    if((mouse_button == GLUT_LEFT_BUTTON) && (mouse_state==GLUT_DOWN)) {
	if(!SELECTED) Find_Nearest(x, y);
	if(SELECTED) {
	    move_ctrlpoint(x, y);
	}
    }
    else
       tbMotion(x, y, mouse_button);
    glutPostRedisplay();
}

void Find_Nearest(int x, int y)
{
    int i, j;
    GLint viewport[4];
    GLdouble mvmatrix[16], projmatrix[16];
    struct POINT p[4][4];
    GLdouble td, dd;

    glGetIntegerv (GL_VIEWPORT, viewport);
    glGetDoublev (GL_MODELVIEW_MATRIX, mvmatrix);
    glGetDoublev (GL_PROJECTION_MATRIX, projmatrix);

  
    for(i=0; i<4; i++) 
	for(j=0; j<4; j++) 
	    p[i][j].x=p[i][j].y=p[i][j].z=0.0;

    for(i=0; i<4; i++) 
	for(j=0; j<4; j++) {
	    p[i][j].x += CtrlPoints[i][j].x * tb_transform[0][0] +
			 CtrlPoints[i][j].y * tb_transform[1][0] +
			 CtrlPoints[i][j].z * tb_transform[2][0] +
			 tb_transform[3][0];
	    p[i][j].y += CtrlPoints[i][j].x * tb_transform[0][1] +
			 CtrlPoints[i][j].y * tb_transform[1][1] +
			 CtrlPoints[i][j].z * tb_transform[2][1] +
			 tb_transform[3][1];
	    p[i][j].z += CtrlPoints[i][j].x * tb_transform[0][2] +
			 CtrlPoints[i][j].y * tb_transform[1][2] +
			 CtrlPoints[i][j].z * tb_transform[2][2] +
			 tb_transform[3][2];
	}
    for(i=0; i<4; i++)
	for(j=0; j<4 ;j++) { 
    	    gluProject((GLdouble)p[i][j].x, 
			(GLdouble)p[i][j].y,
			(GLdouble)p[i][j].z,
			mvmatrix, projmatrix, viewport, 	
			&(wx[i][j]), &(wy[i][j]), &(wz[i][j]));
	    wy[i][j]=(GLdouble)width-wy[i][j];
	    printf("%lf, %lf, %lf\n", wx[i][j], wy[i][j], wz[i][j]);
	}


    printf("\n");    
    printf("x=%d, y=%d\n", x, y); 
    dd=9e+9;

    ni = 0; nj=0;
    for(i=0; i<4; i++)
 	for(j=0;j<4;j++) {
	    td = ((GLdouble)x-wx[i][j]) *  ((GLdouble)x-wx[i][j]) +
		 ((GLdouble)y-wy[i][j]) * ((GLdouble)y-wy[i][j]);
	    if(td<dd) {
		dd=td;
		ni = i;
		nj = j;
	    }
	}
    if(dd<50.) SELECTED = GL_TRUE;
/*
    else SELECTED = GL_FALSE;
*/
    
    
/*
    glColor3f(1.0, 0.0, 0.0);

    glPointSize(4.0);
    glBegin(GL_POINTS);
    glVertex3f(CtrlPoints[i1][j1].x, CtrlPoints[i1][j1].y,
                        CtrlPoints[i1][j1].z);
    glEnd();
*/


}

int main(int argc, char **argv)
{
   glutInit(&argc, argv);
   glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB);
   glutInitWindowSize (500, 500);
   glutInitWindowPosition (100, 100);
   glutCreateWindow (argv[0]);
   init();
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutMouseFunc(mouse);
   glutMotionFunc(motion);

   glutKeyboardFunc (keyboard);
   glutSpecialFunc(special);

   glutMainLoop();
   return 0;
}

static GLboolean invert_matrix( const GLfloat *m, GLfloat *out )
{
/* NB. OpenGL Matrices are COLUMN major. */
#define SWAP_ROWS(a, b) { GLfloat *_tmp = a; (a)=(b); (b)=_tmp; }
#define MAT(m,r,c) (m)[(c)*4+(r)]

 GLfloat wtmp[4][8];
 GLfloat m0, m1, m2, m3, s;
 GLfloat *r0, *r1, *r2, *r3;

 r0 = wtmp[0], r1 = wtmp[1], r2 = wtmp[2], r3 = wtmp[3];

 r0[0] = MAT(m,0,0), r0[1] = MAT(m,0,1),
 r0[2] = MAT(m,0,2), r0[3] = MAT(m,0,3),
 r0[4] = 1.0, r0[5] = r0[6] = r0[7] = 0.0,

 r1[0] = MAT(m,1,0), r1[1] = MAT(m,1,1),
 r1[2] = MAT(m,1,2), r1[3] = MAT(m,1,3),
 r1[5] = 1.0, r1[4] = r1[6] = r1[7] = 0.0,

 r2[0] = MAT(m,2,0), r2[1] = MAT(m,2,1),
 r2[2] = MAT(m,2,2), r2[3] = MAT(m,2,3),
 r2[6] = 1.0, r2[4] = r2[5] = r2[7] = 0.0,

 r3[0] = MAT(m,3,0), r3[1] = MAT(m,3,1),
 r3[2] = MAT(m,3,2), r3[3] = MAT(m,3,3),
 r3[7] = 1.0, r3[4] = r3[5] = r3[6] = 0.0;

 /* choose pivot - or die */
 if (fabs(r3[0])>fabs(r2[0])) SWAP_ROWS(r3, r2);
 if (fabs(r2[0])>fabs(r1[0])) SWAP_ROWS(r2, r1);
 if (fabs(r1[0])>fabs(r0[0])) SWAP_ROWS(r1, r0);
 if (0.0 == r0[0])  return GL_FALSE;

 /* eliminate first variable     */
 m1 = r1[0]/r0[0]; m2 = r2[0]/r0[0]; m3 = r3[0]/r0[0];
 s = r0[1]; r1[1] -= m1 * s; r2[1] -= m2 * s; r3[1] -= m3 * s;
 s = r0[2]; r1[2] -= m1 * s; r2[2] -= m2 * s; r3[2] -= m3 * s;
 s = r0[3]; r1[3] -= m1 * s; r2[3] -= m2 * s; r3[3] -= m3 * s;
 s = r0[4];
 if (s != 0.0) { r1[4] -= m1 * s; r2[4] -= m2 * s; r3[4] -= m3 * s; }
 s = r0[5];
 if (s != 0.0) { r1[5] -= m1 * s; r2[5] -= m2 * s; r3[5] -= m3 * s; }
 s = r0[6];
 if (s != 0.0) { r1[6] -= m1 * s; r2[6] -= m2 * s; r3[6] -= m3 * s; }
 s = r0[7];
 if (s != 0.0) { r1[7] -= m1 * s; r2[7] -= m2 * s; r3[7] -= m3 * s; }

 /* choose pivot - or die */
 if (fabs(r3[1])>fabs(r2[1])) SWAP_ROWS(r3, r2);
 if (fabs(r2[1])>fabs(r1[1])) SWAP_ROWS(r2, r1);
 if (0.0 == r1[1])  return GL_FALSE;

 /* eliminate second variable */
 m2 = r2[1]/r1[1]; m3 = r3[1]/r1[1];
 r2[2] -= m2 * r1[2]; r3[2] -= m3 * r1[2];
 r2[3] -= m2 * r1[3]; r3[3] -= m3 * r1[3];
 s = r1[4]; if (0.0 != s) { r2[4] -= m2 * s; r3[4] -= m3 * s; }
 s = r1[5]; if (0.0 != s) { r2[5] -= m2 * s; r3[5] -= m3 * s; }
 s = r1[6]; if (0.0 != s) { r2[6] -= m2 * s; r3[6] -= m3 * s; }
 s = r1[7]; if (0.0 != s) { r2[7] -= m2 * s; r3[7] -= m3 * s; }

 /* choose pivot - or die */
 if (fabs(r3[2])>fabs(r2[2])) SWAP_ROWS(r3, r2);
 if (0.0 == r2[2])  return GL_FALSE;

 /* eliminate third variable */
 m3 = r3[2]/r2[2];
 r3[3] -= m3 * r2[3], r3[4] -= m3 * r2[4],
 r3[5] -= m3 * r2[5], r3[6] -= m3 * r2[6],
 r3[7] -= m3 * r2[7];

 /* last check */
 if (0.0 == r3[3]) return GL_FALSE;

 s = 1.0/r3[3];              /* now back substitute row 3 */
 r3[4] *= s; r3[5] *= s; r3[6] *= s; r3[7] *= s;

 m2 = r2[3];                 /* now back substitute row 2 */
 s  = 1.0/r2[2];
 r2[4] = s * (r2[4] - r3[4] * m2), r2[5] = s * (r2[5] - r3[5] * m2),
 r2[6] = s * (r2[6] - r3[6] * m2), r2[7] = s * (r2[7] - r3[7] * m2);
 m1 = r1[3];
 r1[4] -= r3[4] * m1, r1[5] -= r3[5] * m1,
 r1[6] -= r3[6] * m1, r1[7] -= r3[7] * m1;
 m0 = r0[3];
 r0[4] -= r3[4] * m0, r0[5] -= r3[5] * m0,
 r0[6] -= r3[6] * m0, r0[7] -= r3[7] * m0;

 m1 = r1[2];                 /* now back substitute row 1 */
 s  = 1.0/r1[1];
 r1[4] = s * (r1[4] - r2[4] * m1), r1[5] = s * (r1[5] - r2[5] * m1),
 r1[6] = s * (r1[6] - r2[6] * m1), r1[7] = s * (r1[7] - r2[7] * m1);
 m0 = r0[2];
 r0[4] -= r2[4] * m0, r0[5] -= r2[5] * m0,
 r0[6] -= r2[6] * m0, r0[7] -= r2[7] * m0;

 m0 = r0[1];                 /* now back substitute row 0 */
 s  = 1.0/r0[0];
 r0[4] = s * (r0[4] - r1[4] * m0), r0[5] = s * (r0[5] - r1[5] * m0),
 r0[6] = s * (r0[6] - r1[6] * m0), r0[7] = s * (r0[7] - r1[7] * m0);

 MAT(out,0,0) = r0[4]; MAT(out,0,1) = r0[5],
 MAT(out,0,2) = r0[6]; MAT(out,0,3) = r0[7],
 MAT(out,1,0) = r1[4]; MAT(out,1,1) = r1[5],
 MAT(out,1,2) = r1[6]; MAT(out,1,3) = r1[7],
 MAT(out,2,0) = r2[4]; MAT(out,2,1) = r2[5],
 MAT(out,2,2) = r2[6]; MAT(out,2,3) = r2[7],
 MAT(out,3,0) = r3[4]; MAT(out,3,1) = r3[5],
 MAT(out,3,2) = r3[6]; MAT(out,3,3) = r3[7];

 return GL_TRUE;

#undef MAT
#undef SWAP_ROWS
}

